/*

  This file is a part of GRALEEDITOR, a program to help create input data
  for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "editorwidget.h"
#include "editorscene.h"
#include "layeritem.h"
#include "pointitem.h"
#include "triangleitem.h"
#include "pointmatchitem.h"
#include <stdio.h>
#include <QWheelEvent>
#include <QMessageBox>
#include <QBrush>
#include <cmath>

EditorWidget::EditorWidget(QWidget *pParent) : QGraphicsView(pParent)
{
	setViewportUpdateMode(BoundingRectViewportUpdate);
	setRenderHint(QPainter::Antialiasing);
	setDragMode(QGraphicsView::RubberBandDrag);

	setBackgroundBrush(Qt::black);

	scale(8.0, 8.0);
	setMatrix(QMatrix(-1,0,0,-1,0,0), true);
}

EditorWidget::~EditorWidget()
{
}

void EditorWidget::wheelEvent(QWheelEvent *event)
{
	scaleView(std::pow((double)2, -event->delta() / 240.0));

	QPoint point = event->pos();

	centerOn(mapToScene(point));
}

void EditorWidget::scaleView(qreal scaleFactor)
{
	qreal factor = matrix().scale(scaleFactor, scaleFactor).mapRect(QRectF(0, 0, 1, 1)).width();

	if (factor < 0.01 || factor > 10000)
	        return;

	scale(scaleFactor, scaleFactor);
}

void EditorWidget::keyPressEvent(QKeyEvent *event)
{
	if (event->key() == Qt::Key_Delete)
	{
		event->accept();
	}
}

void EditorWidget::keyReleaseEvent(QKeyEvent *event)
{
	if (event->key() == Qt::Key_Delete)
	{
		event->accept();

		EditorScene *pScene = (EditorScene *)scene();
		int idx = pScene->getCurrentLayer();
		LayerItem *pLayer = pScene->getLayerItem(idx);

		if (pLayer != 0)
		{
			if (pLayer->getLayerType() == LayerItem::Points)
			{
				QList<PointItem *> points = pLayer->getSelectedPoints();
				QList<TriangleItem *> triangles = pLayer->getSelectedTriangles();

				if (points.size() > 0 || triangles.size() > 0)
				{
					QMessageBox::StandardButton result;

					result = QMessageBox::question(this, "Confirm", "Delete selected points and triangles? Note that additional triangles may need to be deleted.", QMessageBox::Yes|QMessageBox::No, QMessageBox::No);

					if (result == QMessageBox::Yes)
					{
						for (int i = 0 ; i < triangles.size() ; i++)
							delete triangles.at(i);

						for (int i = 0 ; i < points.size() ; i++)
						{
							PointItem *pPoint = points.at(i);

							triangles = pPoint->getTriangles();
							for (int i = 0 ; i < triangles.size() ; i++)
								delete triangles.at(i);
							delete pPoint;
						}
					}
				}
			}
			else
			{
				QList<PointMatchItem *> points = pLayer->getSelectedMatchPoints();

				if (points.size() > 0)
				{
					QMessageBox::StandardButton result;

					result = QMessageBox::question(this, "Confirm", "Delete selected match points?", QMessageBox::Yes|QMessageBox::No, QMessageBox::No);

					if (result == QMessageBox::Yes)
					{
						for (int i = 0 ; i < points.size() ; i++)
						{
							PointMatchItem *pPoint = points.at(i);

							delete pPoint;
						}
					}
				}

			}
		}
	}
}

void EditorWidget::setXAxisOrientation(bool left)
{
	QMatrix m = matrix();

	if (left)
	{
		if (m.m11() < 0)
			return;
		else
			setMatrix(QMatrix(-m.m11(), 0, 0, m.m22(), 0, 0), false);
	}
	else // right
	{
		if (m.m11() > 0)
			return;
		else
			setMatrix(QMatrix(-m.m11(), 0, 0, m.m22(), 0, 0), false);
	}
}

