/*

  This file is a part of GRALEEDITOR, a program to help create input data
  for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "imageitem.h"
#include "fitsitem.h"
#include "pointmatchitem.h"
#include "graleeditor.h"
#include "layeritem.h"
#include "matchpointdialog.h"
#include <QPainter>
#include <QImage>
#include <QPixmap>
#include <QGraphicsSceneMouseEvent>
#include <QMessageBox>
#include <grale/realtype.h>
#include <gsl/gsl_linalg.h>
#include <map>

#include <iostream>

using namespace grale;

ImageItem::ImageItem(QGraphicsItem *pParent) : QGraphicsItem(pParent)
{
	m_pPixmap = 0;
	setCacheMode(QGraphicsItem::DeviceCoordinateCache);
//	setFlag(ItemIsSelectable);
}

ImageItem::~ImageItem()
{
	if (m_pPixmap)
		delete m_pPixmap;
}

bool ImageItem::load(const QString &fname)
{
	QImage img;

	if (!img.load(fname))
		return false;

	clear();

	img = img.mirrored(false, true);
	m_pPixmap = new QPixmap();
	*m_pPixmap = QPixmap::fromImage(img);

	m_imageFileName = fname;
	m_boundingRect = QRectF(QPointF(0, 0), QPointF(m_pPixmap->width(), m_pPixmap->height()));

	setPos(-((double)m_pPixmap->width())*0.05/2.0, ((double)m_pPixmap->height())*0.05/2.0);
	scale(0.05, -0.05);
	updatePointTransforms();

	return true;
}

void ImageItem::clear()
{
	if (m_pPixmap)
	{
		delete m_pPixmap;
		m_pPixmap = 0;
	}
	m_imageFileName = QString();
}

void ImageItem::matchFITSItem(FITSItem *pFITSItem)
{
	if (!pFITSItem)
		return;
	if (!m_pPixmap)
		return;
	if (!pFITSItem->isImageLoaded())
		return;

	LayerItem *pThisLayer = (LayerItem *)parentItem();
	LayerItem *pFITSLayer = (LayerItem *)pFITSItem->parentItem();
	QList<PointMatchItem *> imagePoints = pThisLayer->getAllMatchPoints();
	QList<PointMatchItem *> fitsPoints = pFITSLayer->getAllMatchPoints();

	if (imagePoints.size() == 0 && fitsPoints.size() == 0)
	{
		QRectF fitsBoundary = pFITSItem->boundingRect();
		
		QPointF f1 = pFITSItem->mapToScene(fitsBoundary.x(), fitsBoundary.y());
		QPointF f2 = pFITSItem->mapToScene(fitsBoundary.right(), fitsBoundary.y());
		QPointF f3 = pFITSItem->mapToScene(fitsBoundary.x(), fitsBoundary.bottom());

		Vector2D<double> P1(0, 0);
		Vector2D<double> P2(m_pPixmap->width(), 0);
		Vector2D<double> P3(0, m_pPixmap->height());

		Vector2D<double> p1(f1.x(), f1.y());
		Vector2D<double> p2(f2.x(), f2.y());
		Vector2D<double> p3(f3.x(), f3.y());

		Vector2D<double> d1 = p1-p3;
		Vector2D<double> d2 = p2-p3;

		Vector2D<double> D1 = P1-P3;
		Vector2D<double> D2 = P2-P3;

		double denom = D1.getX()*D2.getY()-D2.getX()*D1.getY();
		double a = 1.0/denom*(d1.getX()*D2.getY()-d2.getX()*D1.getY());
		double b = 1.0/denom*(-d1.getX()*D2.getX()+d2.getX()*D1.getX());
		double c = 1.0/denom*(d1.getY()*D2.getY()-d2.getY()*D1.getY());
		double d = 1.0/denom*(-d1.getY()*D2.getX()+d2.getY()*D1.getX());

		double e = p1.getX()-(a*P1.getX()+b*P1.getY());
		double f = p1.getY()-(c*P1.getX()+d*P1.getY());

		setPos(0, 0);
		setTransform(QTransform(a,c,b,d,e,f), false); // Warning: b and c reversed
		updatePointTransforms();
	}
	else
	{
		std::map<int, std::pair<PointMatchItem *, PointMatchItem *> > pointMap;
		std::map<int, std::pair<PointMatchItem *, PointMatchItem *> >::iterator it;

		for (int i = 0 ; i < imagePoints.size() ; i++)
		{
			PointMatchItem *pPoint = imagePoints.at(i);
			int group = -1;

			if ((group = pPoint->getGroupNumber()) < 0)
				continue;

			it = pointMap.find(group);

			if (it != pointMap.end())
			{
				QMessageBox::warning(GraleEditor::getInstance(), "Error", "Duplicate group ID in image");
				return;
			}

			pointMap[group] = std::pair<PointMatchItem *, PointMatchItem *>(pPoint, 0);
		}

		for (int i = 0 ; i < fitsPoints.size() ; i++)
		{
			PointMatchItem *pPoint = fitsPoints.at(i);
			int group = -1;

			if ((group = pPoint->getGroupNumber()) < 0)
				continue;

			it = pointMap.find(group);

			if (it == pointMap.end())
			{
				QMessageBox::warning(GraleEditor::getInstance(), "Error", "Found group ID in FITS image that isn't present in this image");
				return;
			}

			if ((*it).second.second != 0)
			{
				QMessageBox::warning(GraleEditor::getInstance(), "Error", "Duplicate group ID in FITS image");
				return;
			}

			(*it).second.second = pPoint;
		}

		int numGroups = pointMap.size();

		if (numGroups < 3)
		{
			QMessageBox::warning(GraleEditor::getInstance(), "Error", "At least three point groups are needed");
			return;
		}

		// Check if all entries have two points and store coordinates

		std::vector<Vector2D<double> > imageVectors(numGroups), realCoordinates(numGroups);
		int i = 0;

		for (it = pointMap.begin() ; it != pointMap.end() ; it++, i++)
		{
			std::pair<PointMatchItem *, PointMatchItem *> pointPair = (*it).second;
			PointMatchItem *pImagePoint = pointPair.first;
			PointMatchItem *pFITSPoint = pointPair.second;

			if (pFITSPoint == 0)
			{
				QMessageBox::warning(GraleEditor::getInstance(), "Error", "Found group ID in this image that isn't present in FITS image");
				return;
			}

			QPointF imgPoint = pImagePoint->pos();
			QPointF fitsPoint = pFITSPoint->scenePos();

			imageVectors[i] = Vector2D<double>(imgPoint.x(), imgPoint.y());
			realCoordinates[i] = Vector2D<double>(fitsPoint.x(), fitsPoint.y());
		}

		// find scale factors

		double Scale = findScaleFactor(realCoordinates);
		double scale = findScaleFactor(imageVectors);

		// Find transform
		
		double x2sum = 0;
		double xysum = 0;
		double y2sum = 0;
		double xsum = 0;
		double ysum = 0;
		double Xxsum = 0;
		double Xysum = 0;
		double Yxsum = 0;
		double Yysum = 0;
		double Xsum = 0;
		double Ysum = 0;
		double N = (double)numGroups;

		for (i = 0 ; i < numGroups ; i++)
		{
			double x = imageVectors[i].getX()/scale;
			double y = imageVectors[i].getY()/scale;
			double X = realCoordinates[i].getX()/Scale;
			double Y = realCoordinates[i].getY()/Scale;

			x2sum += x*x;
			xysum += x*y;
			y2sum += y*y;
			xsum += x;
			ysum += y;
			Xxsum += X*x;
			Xysum += X*y;
			Yxsum += Y*x;
			Yysum += Y*y;
			Xsum += X;
			Ysum += Y;
		}

		double matrixElements[36] = { x2sum, xysum, 0, 0, xsum, 0,
		                              xysum, y2sum, 0, 0, ysum, 0,
				              0, 0, x2sum, xysum, 0, xsum,
				              0, 0, xysum, y2sum, 0, ysum,
				              xsum, ysum, 0, 0, N, 0,
				              0, 0, xsum, ysum, 0, N };
		double vectorElements[6] = { Xxsum, Xysum, Yxsum, Yysum, Xsum, Ysum };		

		gsl_matrix_view M = gsl_matrix_view_array(matrixElements, 6, 6);
		gsl_vector_view B = gsl_vector_view_array(vectorElements, 6);
		gsl_vector *pA = gsl_vector_alloc(6);
		gsl_permutation *pPerm = gsl_permutation_alloc(6);

		int s, status;
		double a, b, c, d, e, f;

		if ((status = gsl_linalg_LU_decomp(&(M.matrix), pPerm, &s)) == GSL_SUCCESS)
		{
			if ((status = gsl_linalg_LU_solve(&(M.matrix), pPerm, &(B.vector), pA)) == GSL_SUCCESS)
			{
				a = gsl_vector_get(pA, 0) * (Scale/scale);
				b = gsl_vector_get(pA, 1) * (Scale/scale);
				c = gsl_vector_get(pA, 2) * (Scale/scale);
				d = gsl_vector_get(pA, 3) * (Scale/scale);
				e = gsl_vector_get(pA, 4) * Scale;
				f = gsl_vector_get(pA, 5) * Scale;
			}
		}

		gsl_permutation_free(pPerm);
		gsl_vector_free(pA);

		if (status == GSL_SUCCESS)
		{
			setPos(0, 0);

			QTransform t(a,c,b,d,e,f); // Warning: b and c reversed!
			setTransform(t, false);
			updatePointTransforms();
		}
		else
		{
			QMessageBox::warning(GraleEditor::getInstance(), "Error", "Couldn't find a suitable transformation");
			return;
		}
	}
}

QRectF ImageItem::boundingRect() const
{
	return m_boundingRect;
}

void ImageItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
	if (m_pPixmap == 0)
		return;

	painter->drawPixmap(QPointF(0, 0), *m_pPixmap);

//	painter->setPen(Qt::blue);
//	painter->drawRect(QRectF(QPointF(0, 0), QPointF(m_pPixmap->width(), m_pPixmap->height())));
}

void ImageItem::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
	QGraphicsItem::mousePressEvent(event);
}

void ImageItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
	if (GraleEditor::getInstance()->getCurrentTool() == GraleEditor::PointMatchTool)
	{
		double x = event->pos().x();
		double y = event->pos().y();

		PointMatchItem *pPoint = new PointMatchItem(this);
		LayerItem *pLayer = (LayerItem *)parentItem();

		pPoint->setPos(x, y);
		pPoint->setScaleFactor(pLayer->getPointScaleFactor());
	}
	QGraphicsItem::mouseReleaseEvent(event);
}

double ImageItem::findScaleFactor(const std::vector<Vector2D<double> > points)
{
	double max = 0;

	for (int i = 0 ; i < points.size() ; i++)
	{
		max = MAX(ABS(points[i].getX()), max);
		max = MAX(ABS(points[i].getY()), max);
	}

	return max/2.0;
}

void ImageItem::updatePointTransforms()
{
	LayerItem *pLayer = (LayerItem *)parentItem();

	if (pLayer)
		pLayer->setPointScaleFactor(pLayer->getPointScaleFactor());
}

