/*

  This file is a part of GRALEEDITOR, a program to help create input data
  for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "imagelayertreeitem.h"
#include "fitsitem.h"
#include "imageitem.h"
#include "layeritem.h"
#include <grale/constants.h>
#include <QFileDialog>
#include <QMessageBox>
#include <QList>

#include <iostream>

ImageLayerTreeItem::ImageLayerTreeItem(LayerItem *pLayer, QTreeWidget *pParent) : LayerTreeItem(pLayer, pParent)
{
	setText(0, "Image layer");
	
	TreeItemBase *pLoadItem = new TreeItemBase(TreeItemBase::LoadImage, this);
	pLoadItem->setText(0, "Load image");
	pLoadItem->setFlags(Qt::ItemIsEnabled);
}

ImageLayerTreeItem::~ImageLayerTreeItem()
{
}

void ImageLayerTreeItem::loadImage()
{
	QString fileName = QFileDialog::getOpenFileName(treeWidget(), "Open Image", ".", "FITS Files and image files (*.fits *.fit *.png *.jpg)");

	if (fileName.isNull())
		return;

	FITSItem *pFITSItem = new FITSItem(0);
	ImageItem *pImageItem = new ImageItem(0);
	QGraphicsItem *pGraphicsItem = 0;

	if (pFITSItem->load(fileName))
	{
		delete pImageItem;

		pFITSItem->setParentItem(m_pLayer);
	}
	else if (pImageItem->load(fileName))
	{
		delete pFITSItem;

		pImageItem->setParentItem(m_pLayer);
	}
	else
	{
		QMessageBox::warning(treeWidget(), "Error", "Couldn't load specified file");
		delete pFITSItem;
		delete pImageItem;
		return;
	}
	initializeProperties();
}

void ImageLayerTreeItem::updateImageProperties()
{
	FITSItem *pFITSItem = getFITSItem();

	grale::Vector2D<double> center;
	double min, max;

	if (pFITSItem != 0)
	{
		center = pFITSItem->getCenterPoint();
		center /= CONST_PI;
		center *= 180.0;

		pFITSItem->getMinMax(min, max);
	}	

	int numChildren = childCount();

	double scaleFactor = m_pLayer->getPointScaleFactor();
	int nextGroupNumber = m_pLayer->getNextGroupNumber();

	QString label;
	char str[1024];

	for (int i = 0 ; i < numChildren ; i++)
	{
		TreeItemBase *pProp = (TreeItemBase *)child(i);
		label = QString();

		if (pProp->getType() == TreeItemBase::XCenter)
		{
			label = "X Center";
			sprintf(str, "%10.10f", (double)center.getX());
		}
		else if (pProp->getType() == TreeItemBase::YCenter)
		{
			label = "Y Center";
			sprintf(str, "%10.10f", (double)center.getY());
		}
		else if (pProp->getType() == TreeItemBase::MinValue)
		{
			label = "Min";
			sprintf(str, "%10.10f", (double)min);
		}
		else if (pProp->getType() == TreeItemBase::MaxValue)
		{
			label = "Max";
			sprintf(str, "%10.10f", (double)max);
		}
		else if (pProp->getType() == TreeItemBase::PointScale)
		{
			label = "Point scale factor";
			sprintf(str, "%10.10f", (double)scaleFactor);
		}
		else if (pProp->getType() == TreeItemBase::GroupNumber)
		{
			label = "Next group number";
			sprintf(str, "%d", nextGroupNumber);
		}

		if (!label.isNull())
		{
			pProp->setText(0, label);
			pProp->setText(1, str);
		}
	}
}

FITSItem *ImageLayerTreeItem::getFITSItem()
{
	QList<QGraphicsItem *> children = m_pLayer->childItems();
	FITSItem *pFITSItem = 0;

	for (int i = 0 ; pFITSItem == 0 && i < children.size() ; i++)
		pFITSItem = dynamic_cast<FITSItem *>(children.at(i));

	return pFITSItem;
}

ImageItem *ImageLayerTreeItem::getImageItem()
{
	QList<QGraphicsItem *> children = m_pLayer->childItems();
	ImageItem *pImageItem = 0;

	for (int i = 0 ; pImageItem == 0 && i < children.size() ; i++)
		pImageItem = dynamic_cast<ImageItem *>(children.at(i));

	return pImageItem;
}

void ImageLayerTreeItem::initializeProperties()
{
	FITSItem *pFITSItem = getFITSItem();
	ImageItem *pImageItem = getImageItem();

	if (pFITSItem)
	{
		// Clear the "load image" item
		
		QList<QTreeWidgetItem *> children = takeChildren();

		for (int i = 0 ; i < children.size() ; i++)
			delete children.at(i);

		TreeItemBase *pProp;

		pProp = new TreeItemBase(TreeItemBase::XCenter, this);
		pProp->setFlags(Qt::ItemIsEditable|Qt::ItemIsEnabled);

		pProp = new TreeItemBase(TreeItemBase::YCenter, this);
		pProp->setFlags(Qt::ItemIsEditable|Qt::ItemIsEnabled);

		pProp = new TreeItemBase(TreeItemBase::MinValue, this);
		pProp->setFlags(Qt::ItemIsEditable|Qt::ItemIsEnabled);

		pProp = new TreeItemBase(TreeItemBase::MaxValue, this);
		pProp->setFlags(Qt::ItemIsEditable|Qt::ItemIsEnabled);

		pProp = new TreeItemBase(TreeItemBase::PointScale, this);
		pProp->setFlags(Qt::ItemIsEditable|Qt::ItemIsEnabled);

		pProp = new TreeItemBase(TreeItemBase::GroupNumber, this);
		pProp->setFlags(Qt::ItemIsEditable|Qt::ItemIsEnabled);

		updateImageProperties();

		connect(pFITSItem, SIGNAL(centerChanged()), this, SLOT(updateImageProperties()));
	}
	else if (pImageItem)
	{
		// Clear the "load image" item
		
		QList<QTreeWidgetItem *> children = takeChildren();

		for (int i = 0 ; i < children.size() ; i++)
			delete children.at(i);

		TreeItemBase *pProp;

		pProp = new TreeItemBase(TreeItemBase::MatchFITS, this);
		pProp->setFlags(Qt::ItemIsEnabled);
		pProp->setText(0, "Match FITS image");

		pProp = new TreeItemBase(TreeItemBase::PointScale, this);
		pProp->setFlags(Qt::ItemIsEditable|Qt::ItemIsEnabled);

		pProp = new TreeItemBase(TreeItemBase::GroupNumber, this);
		pProp->setFlags(Qt::ItemIsEditable|Qt::ItemIsEnabled);
		
		updateImageProperties();
	}
}
